# Copyright (C) 2021 FriendliAI

"""PeriFlow Auth Tools"""

import functools
from enum import Enum
from typing import Callable, Union

import requests

from pfcli.utils import get_periflow_directory, get_uri, secho_error_and_exit


access_token_path = get_periflow_directory() / "access_token"
refresh_token_path = get_periflow_directory() / "refresh_token"


class TokenType(str, Enum):
    ACCESS = "ACCESS"
    REFRESH = "REFRESH"


def get_auth_header() -> dict:
    return {"Authorization": f"Bearer {get_token(TokenType.ACCESS)}"}


def get_token(token_type: TokenType) -> Union[str, None]:
    try:
        if token_type == TokenType.ACCESS:
            return access_token_path.read_text()
        if token_type == TokenType.REFRESH:
            return refresh_token_path.read_text()
        else:
            secho_error_and_exit("token_type should be one of 'access' or 'refresh'.")
    except FileNotFoundError:
        return None


def update_token(token_type: TokenType, token: str) -> None:
    if token_type == TokenType.ACCESS:
        access_token_path.write_text(token)
    elif token_type == TokenType.REFRESH:
        refresh_token_path.write_text(token)


def auto_token_refresh(func: Callable[..., requests.Response]) -> Callable[..., requests.Response]:
    @functools.wraps(func)
    def inner(*args, **kwargs) -> requests.Response:
        r = func(*args, **kwargs)
        if r.status_code == 401 or r.status_code == 403:
            refresh_token = get_token(TokenType.REFRESH)
            if refresh_token is not None:
                refresh_r = requests.post(get_uri("token/refresh/"), data={"refresh_token": refresh_token})
                try:
                    refresh_r.raise_for_status()
                except requests.HTTPError:
                    secho_error_and_exit("Failed to refresh access token... Please login again")

                update_token(token_type=TokenType.ACCESS, token=refresh_r.json()["access_token"])
                update_token(token_type=TokenType.REFRESH, token=refresh_r.json()["refresh_token"])
                # We need to restore file offset if we want to transfer file objects
                if "files" in kwargs:
                    files = kwargs["files"]
                    for _, file_tuple in files.items():
                        for element in file_tuple:
                            if hasattr(element, "seek"):
                                # Restore file offset
                                element.seek(0)
                r = func(*args, **kwargs)
                r.raise_for_status()
            else:
                secho_error_and_exit("Failed to refresh access token... Please login again")
        else:
            r.raise_for_status()
        return r
    return inner



# Copyright (C) 2022 FriendliAI

from string import Template
from typing import Dict, Tuple, Type, TypeVar

from pfcli.service import ServiceType
from pfcli.service.client.base import ClientService
from pfcli.service.client.billing import BillingClientService
from pfcli.service.client.checkpoint import CheckpointClientService
from pfcli.service.client.credential import CredentialClientService, CredentialTypeClientService
from pfcli.service.client.data import DataClientService
from pfcli.service.client.experiment import ExperimentClientService
from pfcli.service.client.group import (
    GroupClientService,
    GroupProjectCheckpointClientService,
    GroupProjectClientService,
    GroupVMConfigClientService
)
from pfcli.service.client.job import (
    JobArtifactClientService,
    JobCheckpointClientService,
    JobClientService,
    JobTemplateClientService,
    JobWebSocketClientService
)
from pfcli.service.client.project import (
    ProjectClientService,
    ProjectCredentialClientService,
    ProjectDataClientService,
    ProjectExperimentClientService,
    ProjectJobClientService,
    ProjectVMConfigClientService,
    ProjectVMQuotaClientService
)
from pfcli.service.client.serve import ServeClientService
from pfcli.service.client.user import (
    UserClientService,
    UserGroupClientService,
    UserGroupProjectClientService,
    UserSignUpService
)
from pfcli.utils import get_auth_uri, get_meter_uri, get_mr_uri, get_pfs_uri, get_uri, get_wss_uri

client_template_map: Dict[ServiceType, Tuple[Type[ClientService], Template]] = {
    ServiceType.SIGNUP: (UserSignUpService, Template(get_auth_uri('pf_user/self_signup'))),
    ServiceType.USER: (UserClientService, Template(get_auth_uri('pf_user'))),
    ServiceType.USER_GROUP: (UserGroupClientService, Template(get_auth_uri('pf_user/$pf_user_id/pf_group'))),
    ServiceType.USER_GROUP_PROJECT: (UserGroupProjectClientService, Template(get_auth_uri('pf_user/$pf_user_id/pf_group/$pf_group_id/pf_project'))),  # pylint: disable=line-too-long
    ServiceType.PROJECT: (ProjectClientService, Template(get_auth_uri('pf_project'))),
    ServiceType.GROUP: (GroupClientService, Template(get_auth_uri('pf_group'))),
    ServiceType.GROUP_PROJECT: (GroupProjectClientService, Template(get_auth_uri('pf_group/$pf_group_id/pf_project'))),
    ServiceType.EXPERIMENT: (ExperimentClientService, Template(get_uri('experiment/'))),
    ServiceType.PROJECT_EXPERIMENT: (ProjectExperimentClientService, Template(get_uri('project/$project_id/experiment/'))),  # pylint: disable=line-too-long
    ServiceType.JOB: (JobClientService, Template(get_uri('job/'))),
    ServiceType.JOB_CHECKPOINT: (JobCheckpointClientService, Template(get_uri('job/$job_id/checkpoint/'))),
    ServiceType.JOB_ARTIFACT: (JobArtifactClientService, Template(get_uri('job/$job_id/artifact/'))),
    ServiceType.PROJECT_JOB: (ProjectJobClientService, Template(get_uri('project/$project_id/job/'))),
    ServiceType.JOB_TEMPLATE: (JobTemplateClientService, Template(get_uri('job_template/'))),
    ServiceType.CREDENTIAL: (CredentialClientService, Template(get_auth_uri('credential'))),
    ServiceType.PROJECT_CREDENTIAL: (ProjectCredentialClientService, Template(get_auth_uri('pf_project/$project_id/credential'))),  # pylint: disable=line-too-long
    ServiceType.CREDENTIAL_TYPE: (CredentialTypeClientService, Template(get_uri('credential_type/'))),
    ServiceType.DATA: (DataClientService, Template(get_uri('datastore/'))),
    ServiceType.PROJECT_DATA: (ProjectDataClientService, Template(get_uri('project/$project_id/datastore/'))),
    ServiceType.PROJECT_VM_QUOTA: (ProjectVMQuotaClientService, Template(get_uri('project/$project_id/vm_quota/'))),
    ServiceType.CHECKPOINT: (CheckpointClientService, Template(get_mr_uri('models/'))),
    ServiceType.GROUP_PROJECT_CHECKPOINT: (GroupProjectCheckpointClientService, Template(get_mr_uri('orgs/$group_id/prjs/$project_id/models/'))),  # pylint: disable=line-too-long
    ServiceType.PROJECT_VM_CONFIG: (ProjectVMConfigClientService, Template(get_uri('project/$project_id/vm_config/'))),
    ServiceType.GROUP_VM_CONFIG: (GroupVMConfigClientService, Template(get_uri('group/$group_id/vm_config/'))),
    ServiceType.JOB_WS: (JobWebSocketClientService, Template(get_wss_uri('job/'))),
    ServiceType.SERVE: (ServeClientService, Template(get_pfs_uri('deployment/'))),
    ServiceType.BILLING_SUMMARY: (BillingClientService, Template(get_meter_uri('training/instances/price/')))
}


_ClientService = TypeVar('_ClientService', bound=ClientService)


def build_client(request_type: ServiceType, **kwargs) -> _ClientService:
    """Factory function to post client service.

    Args:
        request_type (RequestAPI):

    Returns:
        ClientService: created client service
    """
    cls, template = client_template_map[request_type]
    return cls(template, **kwargs)



# Copyright (C) 2021 FriendliAI

from collections import defaultdict
from functools import reduce
from typing import Optional

import tabulate
import typer

from pfcli.service import ServiceType
from pfcli.service.client import build_client
from pfcli.service.client.billing import BillingClientService
from pfcli.context import get_current_group_id, get_current_project_id
from pfcli.service.formatter import PanelFormatter, TableFormatter
from pfcli.utils import secho_error_and_exit


tabulate.PRESERVE_WHITESPACE = True

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False
)

table_formatter = TableFormatter(
    name="Billing",
    fields = ['vm_name', 'price'],
    headers = ['VM Name', 'Amount used in USD']
)

panel_formatter = PanelFormatter(
    name="Total",
    fields = ['price'],
    headers = ['Total amount used in USD']
)

@app.command(help="summarize billing information")
def summary(
    year: int = typer.Argument(
        ...
    ),
    month: int = typer.Argument(
        ...
    ),
    day: Optional[int] = typer.Argument(
        None
    ),
    view_organization: bool = typer.Option(
        False,
        '--organization',
        '-o',
        help='View organization-level cost summary'
    )
):
    "Summarize the billing information for the given time range"
    client: BillingClientService = build_client(ServiceType.BILLING_SUMMARY)
    group_id = None
    project_id = None
    if not view_organization:
        project_id = get_current_project_id()
        if project_id is None:
            secho_error_and_exit("Project is not set!")
    else:
        group_id = get_current_group_id()
        if group_id is None:
            secho_error_and_exit("Organization is not set!")

    prices = client.list_prices(year=year,
                                month=month,
                                day=day,
                                group_id=group_id,
                                project_id=project_id)

    price_by_vmname = defaultdict(float)

    for price_info in prices:
        price_by_vmname[price_info['instance']['attributes']['vm_name']] += price_info['price']
    table_formatter.render([{"vm_name": vm_name, "price": round(price, 2)}
                            for vm_name,price in price_by_vmname.items()])

    price_sum = reduce((lambda acc, x: acc + x), price_by_vmname.values(), 0.)

    total_price = [{"price": round(price_sum, 2)}]
    panel_formatter.render(total_price)



# Copyright (C) 2021 FriendliAI

"""CLI for Checkpoint"""

import os
from dateutil.parser import parse
from typing import Optional, List

import typer

from pfcli.service import (
    CheckpointCategory,
    ModelFormCategory,
    ServiceType,
    StorageType,
    cred_type_map,
    cred_type_map_inv,
    storage_type_map_inv,
)
from pfcli.service.client import (
    CheckpointClientService,
    CredentialClientService,
    GroupProjectCheckpointClientService,
    build_client,
)
from pfcli.service.cloud import build_storage_helper
from pfcli.service.formatter import PanelFormatter, TableFormatter, TreeFormatter
from pfcli.utils import datetime_to_pretty_str, download_file, secho_error_and_exit


app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False
)

table_formatter = TableFormatter(
    name="Checkpoints",
    fields=['id', 'name', 'model_category', 'forms[0].vendor', 'forms[0].storage_name', 'iteration', 'forms[0].form_category', 'created_at'],
    headers=['ID', 'Name', 'Category', 'Cloud', 'Storage Name', 'Iteration', 'Format', 'Created At']
)
panel_formatter = PanelFormatter(
    name="Overview",
    fields=['id', 'name', 'model_category', 'forms[0].vendor', 'forms[0].storage_name', 'iteration', 'forms[0].form_category', 'created_at'],
    headers=['ID', 'Name', 'Category', 'Cloud', 'Storage Name', 'Iteration', 'Format', 'Created At']
)
tree_formatter = TreeFormatter(name="Files")


def _validate_parallelism_order(value: str) -> List[str]:
    parallelism_order = value.split(",")
    if {"pp", "dp", "mp"} != set(parallelism_order):
        secho_error_and_exit("Invalid Argument: parallelism_order should contain 'pp', 'dp', 'mp'")
    return parallelism_order


@app.command()
def list(
    category: Optional[CheckpointCategory] = typer.Option(
        None,
        "--category",
        "-c",
        help="Category of checkpoints. One of 'user_provided' and 'job_generated'."
    )
):
    """List all checkpoints that belong to the user's organization.
    """
    client: GroupProjectCheckpointClientService = build_client(ServiceType.GROUP_PROJECT_CHECKPOINT)
    checkpoints = client.list_checkpoints(category)
    for ckpt in checkpoints:
        for form in ckpt['forms']:
            form['vendor'] = storage_type_map_inv[form['vendor']].value
        ckpt['created_at'] = datetime_to_pretty_str(parse(ckpt['created_at']))

    table_formatter.render(checkpoints)


@app.command()
def view(
    checkpoint_id: str = typer.Argument(
        ...,
        help="UUID of checkpoint to inspect detail."
    )
):
    """Show details of a checkpoint.
    """
    client: CheckpointClientService = build_client(ServiceType.CHECKPOINT)
    ckpt = client.get_checkpoint(checkpoint_id)
    for form in ckpt['forms']:
        form['vendor'] = storage_type_map_inv[form['vendor']].value
    ckpt['created_at'] = datetime_to_pretty_str(parse(ckpt['created_at']))

    panel_formatter.render([ckpt])
    tree_formatter.render(ckpt['forms'][0]['files'])


@app.command()
def create(
    name: str = typer.Option(
        ...,
        '--name',
        '-n',
        help="Name of your checkpoint to create."
    ),
    format: ModelFormCategory = typer.Option(
        ModelFormCategory.ETC,
        '-m',
        '--format',
        help="The format of your checkpoint",
    ),
    cloud: StorageType = typer.Option(
        ...,
        "--cloud",
        "-c",
        help="The cloud storage vendor type where the checkpoint is uploaded."
    ),
    region: str = typer.Option(
        ...,
        "--region",
        "-r",
        help="The cloud storage region where the checkpoint is uploaded."
    ),
    storage_name: str = typer.Option(
        ...,
        "--storage-name",
        "-s",
        help="The name cloud storage where the checkpoint is uploaded."
    ),
    storage_path: Optional[str] = typer.Option(
        None,
        "--storage-path",
        "-p",
        help="File or direcotry path of cloud storage. The root of the storage will be used by default."
    ),
    credential_id: str = typer.Option(
        ...,
        "--credential-id",
        "-i",
        help="UUID of crendential to access cloud storage."
    ),
    iteration: int = typer.Option(
        ...,
        "--iteration",
        help="The iteration number of the checkpoint."
    ),
    # advanced arguments
    dp_degree: int = typer.Option(
        1,
        "--dp-degree",
        help="Data parallelism degree of the model checkpoint."
    ),
    pp_degree: int = typer.Option(
        1,
        "--pp-degree",
        help="Pipelined model parallelism degree of the model checkpoint."
    ),
    mp_degree: int = typer.Option(
        1,
        "--mp-degree",
        help="Tensor parallelism degree of the model checkpoint."
    ),
    parallelism_order: str = typer.Option(
        'pp,dp,mp',
        '--parallelism-order',
        callback=_validate_parallelism_order,
        help="Order of device allocation in distributed training."
    )
):
    """Create a checkpoint object by registering user's cloud storage to PeriFlow.
    """
    dist_config = {
        "pp_degree": pp_degree,
        "dp_degree": dp_degree,
        "mp_degree": mp_degree,
        "dp_mode": "allreduce",
        "parallelism_order": parallelism_order
    }

    credential_client: CredentialClientService = build_client(ServiceType.CREDENTIAL)
    credential = credential_client.get_credential(credential_id)
    if credential["type"] != cred_type_map[cloud]:
        secho_error_and_exit(
            f"Credential type and cloud vendor mismatch: {cred_type_map_inv[credential['type']]} and {cloud}."
        )

    storage_helper = build_storage_helper(cloud, credential_json=credential["value"])
    if storage_path is not None:
        storage_path = storage_path.strip('/')
    files = storage_helper.list_storage_files(storage_name, storage_path)
    if storage_path is not None:
        storage_name = f"{storage_name}/{storage_path}"

    checkpoint_client: GroupProjectCheckpointClientService = build_client(ServiceType.GROUP_PROJECT_CHECKPOINT)
    ckpt = checkpoint_client.create_checkpoint(
        name=name,
        model_form_category=format,
        vendor=cloud,
        region=region,
        credential_id=credential_id,
        iteration=iteration,
        storage_name=storage_name,
        files=files,
        dist_config=dist_config,
        data_config={},
        job_setting_config=None   # TODO: make configurable
    )
    for form in ckpt['forms']:
        form['vendor'] = storage_type_map_inv[form['vendor']].value
    ckpt['created_at'] = datetime_to_pretty_str(parse(ckpt['created_at']))

    panel_formatter.render([ckpt])
    tree_formatter.render(ckpt['files'])


@app.command()
def delete(
    checkpoint_id: str = typer.Argument(
        ...,
        help="UUID of checkpoint to delete."
    ),
    force: bool = typer.Option(
        False,
        '--force',
        '-f',
        help="Forcefully delete checkpoint without confirmation prompt."
    )
):
    """Delete the existing checkpoint.
    """
    if not force:
        do_delete = typer.confirm("Are you sure to delete checkpoint?")
        if not do_delete:
            raise typer.Abort()

    client: CheckpointClientService = build_client(ServiceType.CHECKPOINT)
    client.delete_checkpoint(checkpoint_id)

    typer.secho("Checkpoint is deleted successfully!", fg=typer.colors.BLUE)


@app.command()
def download(
    checkpoint_id: str = typer.Argument(
        ...,
        help="UUID of checkpoint to download."
    ),
    save_directory: Optional[str] = typer.Option(
        None,
        '--destination',
        '-d',
        help="Destination path to directory to save checkpoint files."
    )
):
    """Download checkpoint files to local storage.
    """
    if save_directory is not None and not os.path.isdir(save_directory):
        secho_error_and_exit(f"Directory {save_directory} is not found.")

    save_directory = save_directory or os.getcwd()

    client: CheckpointClientService = build_client(ServiceType.CHECKPOINT)
    files = client.get_checkpoint_download_urls(checkpoint_id)

    for i, file in enumerate(files):
        typer.secho(f"Downloading files {i + 1}/{len(files)}...")
        download_file(file['download_url'], out=os.path.join(save_directory, file['name']))



# Copyright (C) 2022 FriendliAI

"""PeriFlow ExperimentClient Service"""


from typing import List

from pfcli.service.client.base import ClientService, T, safe_request


class ExperimentClientService(ClientService):
    def list_jobs_in_experiment(self, experiment_id: T) -> List[dict]:
        response = safe_request(self.list, err_prefix="Failed to fetch jobs in the experiment.")(
            path=f"{experiment_id}/job/"
        )
        return response.json()['results']

    def delete_experiment(self, experiment_id: T) -> None:
        safe_request(self.delete, err_prefix="Failed to delete experiment.")(
            pk=experiment_id
        )

    def update_experiment_name(self, experiment_id: T, name: str) -> dict:
        response = safe_request(self.partial_update, err_prefix=f"Failed to update the name of experiment to {name}")(
            pk=experiment_id,
            json={"name": name}
        )
        return response.json()



# Copyright (C) 2021 FriendliAI

"""PeriFlow Credential CLI"""

import json

import typer

from pfcli.service import ServiceType, CredType, cred_type_map_inv
from pfcli.service.client import CredentialClientService, ProjectCredentialClientService, build_client
from pfcli.service.config import CredentialConfigService
from pfcli.service.formatter import PanelFormatter, TableFormatter
from pfcli.utils import secho_error_and_exit


app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False
)
create_app = typer.Typer(
    help="Create credential",
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False
)

app.add_typer(create_app, name='create')

table_formatter = TableFormatter(
    name="Credentials",
    fields=['id', 'name', 'type', 'created_at', 'owner_type'],
    headers=['ID', 'Name', 'Type', 'Created At', 'Scope']
)

panel_formatter = PanelFormatter(
    name="Overview",
    fields=['id', 'name', 'type', 'created_at', 'owner_type'],
    headers=['ID', 'Name', 'Type', 'Created At', 'Scope']
)


S3_DOC_LINK = "https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-quickstart.html"
AZURE_BLOB_DOC_LINK = "https://docs.microsoft.com/en-us/azure/storage/common/storage-account-keys-manage?tabs=azure-portal" # pylint: disable=line-too-long
GCP_DOC_LINK = "https://cloud.google.com/iam/docs/creating-managing-service-account-keys"
SLACK_DOC_LINK = "https://slack.com/help/articles/215770388-Create-and-regenerate-API-tokens"
WANDB_API_KEY_LINK = "https://wandb.ai/authorize"


@create_app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    interactive: bool = typer.Option(
        False,
        '--interactive',
        help="Use interactive mode."
    )
):
    if not interactive:
        if ctx.invoked_subcommand is None:
            secho_error_and_exit(
                f"You should provide one of the credential types {[ e.value for e in CredType ]} as an argument, "
                "or use '--interactive' option. Run 'pf credential --help' for more details."
            )
        return

    configurator = CredentialConfigService()

    configurator.start_interaction()
    name, cred_type, value = configurator.render()

    client: ProjectCredentialClientService = build_client(ServiceType.PROJECT_CREDENTIAL)

    cred = client.create_credential(cred_type, name, 1, value)
    cred['type'] = cred_type_map_inv[cred['type']].value

    typer.secho("Credential created successfully!", fg=typer.colors.BLUE)
    panel_formatter.render([cred])
    exit(0)


@create_app.command()
def docker(
    name: str = typer.Option(
        ...,
        '--name',
        '-n',
        help="Name of the new credential."
    ),
    username: str = typer.Option(
        ...,
        help="Docker username."
    ),
    password: str = typer.Option(
        ...,
        help="Docker password."
    ),
):
    """Create a credential for Docker.
    """
    client: ProjectCredentialClientService = build_client(ServiceType.PROJECT_CREDENTIAL)
    value = {
        'username': username,
        'password': password
    }
    cred = client.create_credential(CredType.DOCKER, name, 1, value)
    cred['type'] = cred_type_map_inv[cred['type']].value
    panel_formatter.render([cred])


@create_app.command()
def s3(
    name: str = typer.Option(
        ...,
        '--name',
        '-n',
        help="Name of the new credential."
    ),
    aws_access_key_id: str = typer.Option(
        ...,
        help=f"[AWS] AWS Acess Key ID. Please see {S3_DOC_LINK}."
    ),
    aws_secret_access_key: str = typer.Option(
        ...,
        help=f"[AWS] AWS Secret Access Key. Please see {S3_DOC_LINK}."
    ),
    aws_default_region: str = typer.Option(
        ...,
        help=f"[AWS] Default region name. Please see {S3_DOC_LINK}."
    ),
):
    """Create a credential for AWS S3.
    """
    client: ProjectCredentialClientService = build_client(ServiceType.PROJECT_CREDENTIAL)
    value = {
        'aws_access_key_id': aws_access_key_id,
        'aws_secret_access_key': aws_secret_access_key,
        'aws_default_region': aws_default_region
    }
    cred = client.create_credential(CredType.S3, name, 1, value)
    cred['type'] = cred_type_map_inv[cred['type']].value
    panel_formatter.render([cred])


@create_app.command()
def azure_blob(
    name: str = typer.Option(
        ...,
        '--name',
        '-n',
        help="Name of the new credential."
    ),
    storage_account_name: str = typer.Option(
        ...,
        help=f"[Azure] Azure Blob storage account name."
    ),
    storage_account_key: str = typer.Option(
        ...,
        help=f"[Azure] Azure Blob storage account access key. Please see {AZURE_BLOB_DOC_LINK}."
    ),
):
    """Create a credential for Azure Blob storage.
    """
    client: ProjectCredentialClientService = build_client(ServiceType.PROJECT_CREDENTIAL)
    value = {
        'storage_account_name': storage_account_name,
        'storage_account_key': storage_account_key,
    }
    cred = client.create_credential(CredType.BLOB, name, 1, value)
    cred['type'] = cred_type_map_inv[cred['type']].value
    panel_formatter.render([cred])


@create_app.command()
def gcs(
    name: str = typer.Option(
        ...,
        '--name',
        '-n',
        help="Name of the new credential."
    ),
    service_account_key_file: typer.FileText = typer.Option(
        ...,
        help=f"[GCP] Path to GCP Service Account Key JSON file. Please see {GCP_DOC_LINK}."
    ),
):
    """Create a credential for Google Cloud Storage.
    """
    client: ProjectCredentialClientService = build_client(ServiceType.PROJECT_CREDENTIAL)
    try:
        value = json.load(service_account_key_file)
    except json.JSONDecodeError as exc:
        secho_error_and_exit(f"Error occurred while parsing JSON file... {exc}")
    del value['type']
    cred = client.create_credential(CredType.GCS, name, 1, value)
    cred['type'] = cred_type_map_inv[cred['type']].value
    panel_formatter.render([cred])


@create_app.command()
def slack(
    name: str = typer.Option(
        ...,
        '--name',
        '-n',
        help="Name of the new credential."
    ),
    token: str = typer.Option(
        ...,
        help=f"Slack API Token. Please see {SLACK_DOC_LINK}."
    ),
):
    """Create a credential for Slack.
    """
    client: ProjectCredentialClientService = build_client(ServiceType.PROJECT_CREDENTIAL)
    value = {
        'token': token
    }
    cred = client.create_credential(CredType.SLACK, name, 1, value)
    cred['type'] = cred_type_map_inv[cred['type']].value
    panel_formatter.render([cred])


@create_app.command()
def wandb(
    name: str = typer.Option(
        ...,
        '--name',
        '-n',
        help="Name of the new credential."
    ),
    api_key: str = typer.Option(
        ...,
        help=f"Weights & Biases API Key. You can get the key from {WANDB_API_KEY_LINK}."
    ),
):
    """Create a credential for Weights & Biases.
    """
    client: ProjectCredentialClientService = build_client(ServiceType.PROJECT_CREDENTIAL)

    value = {
        'token': api_key
    }
    cred = client.create_credential(CredType.WANDB, name, 1, value)
    cred['type'] = cred_type_map_inv[cred['type']].value
    panel_formatter.render([cred])


@app.command()
def list(
    cred_type: CredType = typer.Option(
        ...,
        '--cred-type',
        '-t',
        help="Type of credentials to list."
    ),
):
    """List credentials.
    """
    client: ProjectCredentialClientService = build_client(ServiceType.PROJECT_CREDENTIAL)
    creds = client.list_credentials(cred_type)
    for cred in creds:
        cred['type'] = cred_type_map_inv[cred['type']].value

    table_formatter.render(creds)


@app.command()
def update(
    credential_id: str = typer.Argument(
        ...,
        help='UUID of credential to update.'
    )
):
    """Update a credential data.
    """
    configurator = CredentialConfigService()

    configurator.start_interaction_for_update(credential_id)
    name, _, value = configurator.render()

    client: CredentialClientService = build_client(ServiceType.CREDENTIAL)

    cred = client.update_credential(credential_id, name=name, type_version=1, value=value)
    cred['type'] = cred_type_map_inv[cred['type']].value

    typer.secho("Credential updated successfully!", fg=typer.colors.BLUE)
    panel_formatter.render([cred])


@app.command()
def delete(
    credential_id: str = typer.Argument(
        ...,
        help='UUID of credential to delete.'
    ),
    force: bool = typer.Option(
        False,
        '--force',
        '-f',
        help="Forcefully delete credential without confirmation prompt."
    )
):
    """Delete a credential.
    """
    if not force:
        do_delete = typer.confirm("Are you sure to delete credential?")
        if not do_delete:
            raise typer.Abort()

    client: CredentialClientService = build_client(ServiceType.CREDENTIAL)
    client.delete_credential(credential_id)

    typer.secho(f"Credential ({credential_id}) is deleted successfully!", fg=typer.colors.BLUE)



# Copyright (C) 2022 FriendliAI

"""PeriFlow JobClient Service"""

import json
import requests
from contextlib import asynccontextmanager
from typing import Iterator, List, Optional
from requests.models import Response

import websockets
from websockets.client import WebSocketClientProtocol
from websockets.exceptions import ConnectionClosed

from pfcli.service import LogType
from pfcli.service.auth import TokenType, auto_token_refresh, get_auth_header, get_token
from pfcli.service.client.base import ClientService, safe_request
from pfcli.utils import paginated_get, secho_error_and_exit


class JobClientService(ClientService):
    def list_jobs(self) -> List[dict]:
        return paginated_get(safe_request(self.list, err_prefix="Failed to list jobs."))

    def delete_job(self, job_id: int) -> None:
        safe_request(self.delete, err_prefix=f"Failed to delete job ({job_id}).")(
            pk=job_id
        )

    def get_job(self, job_id: int) -> dict:
        response = safe_request(self.retrieve, err_prefix="Failed to list jobs.")(
            pk=job_id
        )
        return response.json()

    def cancel_job(self, job_id: int) -> None:
        safe_request(self.post, err_prefix=f"Failed to cancel job ({job_id}).")(
            path=f"{job_id}/cancel/"
        )

    def terminate_job(self, job_id: int) -> None:
        safe_request(self.post, err_prefix=f"Failed to terminate job ({job_id}).")(
            path=f"{job_id}/terminate/"
        )

    def get_text_logs(self,
                      job_id: int,
                      num_records: int,
                      head: bool = False,
                      log_types: Optional[List[str]] = None,
                      machines: Optional[List[int]] = None,
                      content: Optional[str] = None) -> List[dict]:
        request_data = {'limit': num_records}
        if head:
            request_data['ascending'] = 'true'
        else:
            request_data['ascending'] = 'false'
        if content is not None:
            request_data['content'] = content
        if log_types is not None:
            request_data['log_types'] = ",".join(log_types)
        if machines is not None:
            request_data['node_ranks'] = ",".join([str(machine) for machine in machines])

        logs = paginated_get(
            safe_request(self.list, err_prefix="Failed to fetch text logs."),
            path=f"{job_id}/text_log/",
            **request_data
        )
        if not head:
            logs.reverse()

        return logs


class JobWebSocketClientService(ClientService):
    @asynccontextmanager
    async def _connect(self, job_id: int) -> Iterator[WebSocketClientProtocol]:
        access_token = get_token(TokenType.ACCESS)
        base_url = self.url_template.render(job_id)
        url = f'{base_url}?token={access_token}'
        async with websockets.connect(url) as websocket:
            yield websocket

    async def _subscribe(self,
                         websocket: WebSocketClientProtocol,
                         sources: List[str],
                         node_ranks: List[int]):
        subscribe_json = {
            "type": "subscribe",
            "sources": sources,
            "node_ranks": node_ranks
        }
        await websocket.send(json.dumps(subscribe_json))
        response = await websocket.recv()
        try:
            decoded_response = json.loads(response)
            assert decoded_response.get("response_type") == "subscribe" and \
                set(sources) == set(decoded_response["sources"])
        except json.JSONDecodeError:
            secho_error_and_exit("Error occurred while decoding websocket response...")
        except AssertionError:
            secho_error_and_exit(f"Invalid websocket response... {response}")

    @asynccontextmanager
    async def open_connection(self,
                              job_id: int,
                              log_types: Optional[List[str]],
                              machines: Optional[List[int]]):
        if log_types is None:
            sources = [f"process.{x.value}" for x in LogType]
        else:
            sources = [f"process.{log_type}" for log_type in log_types]

        if machines is None:
            node_ranks = []
        else:
            node_ranks = machines

        async with self._connect(job_id) as websocket:
            websocket: WebSocketClientProtocol
            await self._subscribe(websocket, sources, node_ranks)
            self._websocket = websocket
            yield

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            response = await self._websocket.recv()
        except ConnectionClosed as exc:
            raise StopAsyncIteration from exc   # pragma: no cover

        try:
            return json.loads(response)
        except json.JSONDecodeError:
            secho_error_and_exit("Error occurred while decoding websocket response...")


class JobCheckpointClientService(ClientService):
    def list_checkpoints(self) -> List[dict]:
        response = safe_request(self.list, err_prefix="Failed to list checkpoints.")()
        return response.json()


class JobArtifactClientService(ClientService):
    def list_artifacts(self) -> List[dict]:
        response = safe_request(self.list, err_prefix="Failed to list artifacts.")()
        return response.json()

    @auto_token_refresh
    def download(self, artifact_id: int) -> Response:
        url_template = self.url_template.copy()
        url_template.attach_pattern(f"{artifact_id}/download/")
        return requests.get(
            url_template.render(**self.url_kwargs),
            headers=get_auth_header()
        )

    def get_artifact_download_url(self, artifact_id: int) -> dict:
        response = safe_request(self.download, err_prefix="Failed to get artifact download url")(artifact_id)
        return response.json()


class JobTemplateClientService(ClientService):
    def list_job_template_names(self) -> List[str]:
        response = safe_request(self.list, err_prefix="Failed to list job template names.")()
        return [ template['name'] for template in response.json() ]

    def get_job_template_by_name(self, name: str) -> Optional[dict]:
        response = safe_request(self.list, err_prefix="Failed to list job template names.")()
        for template in response.json():
            if template['name'] == name:
                return template
        return None



# Copyright (C) 2021 FriendliAI

"""PeriFlow Cloud Service"""

from typing import (
    Callable,
    Dict,
    List,
    TypeVar,
    Type,
    Optional,
    Union,
    Tuple,
)
from dataclasses import dataclass

import boto3
from botocore.exceptions import ClientError
from botocore.client import BaseClient
from azure.storage.blob import BlobServiceClient

from pfcli.service import StorageType
from pfcli.utils import secho_error_and_exit


C = TypeVar('C', bound='CloudStorageHelper')
T = TypeVar('T', bound=Union[BaseClient, BlobServiceClient])


@dataclass
class CloudStorageHelper:
    client: T

    def list_storage_files(self, storage_name: str, path_prefix: Optional[str] = None) -> List[dict]:
        raise NotImplementedError   # pragma: no cover


@dataclass
class AWSCloudStorageHelper(CloudStorageHelper):
    def _check_aws_bucket_exists(self, storage_name: str) -> bool:
        try:
            self.client.head_bucket(Bucket=storage_name)
            return True
        except ClientError:
            # include both Forbidden access, Not Exists
            return False

    def list_storage_files(self, storage_name: str, path_prefix: Optional[str] = None) -> List[dict]:
        if not self._check_aws_bucket_exists(storage_name):
            secho_error_and_exit(f"Bucket {storage_name} does not exist")

        file_list = []
        prefix_option = {'Prefix': path_prefix} if path_prefix is not None else {}
        object_contents = self.client.list_objects(Bucket=storage_name, **prefix_option)['Contents']
        for object_content in object_contents:
            object_key = object_content['Key']
            name = object_key.split('/')[-1]
            if not name:
                continue    # skip directory
            file_list.append({
                'name': name,
                'path': object_key,
                'mtime': object_content['LastModified'].isoformat(),
                'size': object_content['Size']
            })

        if not file_list:
            secho_error_and_exit(f"No file exists in Bucket {storage_name}")

        return file_list


@dataclass
class AzureCloudStorageHelper(CloudStorageHelper):
    def list_storage_files(self, storage_name: str, path_prefix: Optional[str] = None):
        container_client = self.client.get_container_client(storage_name)
        if not container_client.exists():
            secho_error_and_exit(f"Container {storage_name} does not exist")

        file_list = []
        prefix_option = {'name_starts_with': path_prefix} if path_prefix is not None else {}
        object_contents = container_client.list_blobs(**prefix_option)
        for object_content in object_contents:
            object_name = object_content['name']
            name = object_name.split('/')[-1]
            if not name:
                continue    # skip directory
            file_list.append({
                'name': name,
                'path': object_name,
                'mtime': object_content['last_modified'].isoformat(),
                'size': object_content['size'],
            })

        if not file_list:
            secho_error_and_exit(f"No file exists in Bucket {storage_name}")

        return file_list


def build_s3_client(credential_json: Dict[str, str]) -> BaseClient:
    return boto3.client(
        's3',
        aws_access_key_id=credential_json['aws_access_key_id'],
        aws_secret_access_key=credential_json['aws_secret_access_key'],
        region_name=credential_json.get('aws_default_region', None)
    )


def build_blob_client(credential_json: Dict[str, str]) -> BlobServiceClient:
    url = f"https://{credential_json['storage_account_name']}.blob.core.windows.net/"
    return BlobServiceClient(
        account_url=url,
        credential=credential_json['storage_account_key']
    )


# TODO: Add GCP support
vendor_helper_map: Dict[StorageType, Tuple[Type[C], Callable[[Dict[str, str]], T]]] = {
    StorageType.S3: (AWSCloudStorageHelper, build_s3_client),
    StorageType.BLOB: (AzureCloudStorageHelper, build_blob_client),
}


def build_storage_helper(vendor: StorageType, credential_json: dict) -> C:
    cls, client_build_fn = vendor_helper_map[vendor]
    client = client_build_fn(credential_json)
    return cls(client)



# Copyright (C) 2021 FriendliAI

"""PeriFlow CLI Output Formatter"""

import json
import re
from dataclasses import dataclass, field
from typing import (
    Any,
    Dict,
    List,
    Optional,
    TypeVar,
    Union,
)

from rich.table import Table
from rich.panel import Panel
from rich.tree import Tree
from rich.text import Text
from rich.json import JSON
from rich.filesize import decimal
from rich.console import Console, RenderableType


T = TypeVar('T', bound=Union[int, str])


def get_value(data: dict, key: str) -> T:
    """Get value of `key` from `data`.
    Unlike dict.get method, it is available to access the nested value in `data`.

    Args:
        data (dict): Data to read.
        key (str): Dot(.)-separated nested keys in data to access the value.

    Returns:
        T: The retrieved data.
    """
    value: Any = data
    keys = key.split('.')
    for key in keys:
        # e.g. `data.results[2].a`
        getitem_match = re.match(r"(.+)\[(-?\d+)\]$", key)
        if getitem_match:
            value = value.get(getitem_match.group(1))[int(getitem_match.group(2))]
        else:
            value = value.get(key)

    return str(value)


@dataclass
class Formatter:
    """Formatter helps data formatting and visualization.
    """
    name: str

    def __post_init__(self):
        self._console = Console()


@dataclass
class ListFormatter(Formatter):
    """Base interface for list data formatter.
    """
    fields: List[str]
    headers: List[str]
    extra_fields: List[str] = field(default_factory=list)
    extra_headers: List[str] = field(default_factory=list)

    def __post_init__(self):
        super().__post_init__()
        assert len(self.fields) == len(self.headers)
        assert len(self.extra_fields) == len(self.extra_headers)

        self._styling_map: Dict[str, Dict[str, Any]] = {}
        self._substitution_rule: Dict[str, str] = {}

    def render(self, data: List[dict], show_detail: bool = False) -> None:
        raise NotImplementedError   # pragma: no cover

    def get_renderable(self, data: List[dict], show_detail: bool = False) -> RenderableType:
        raise NotImplementedError   # pragma: no cover

    def apply_styling(self, header: str, **kwargs) -> None:
        self._styling_map[header] = kwargs

    def add_substitution_rule(self, before: str, after: str) -> None:
        self._substitution_rule[before] = after

    def _substitute(self, val: str) -> str:
        if val in self._substitution_rule:
            return self._substitution_rule[val]
        return val


@dataclass
class TableFormatter(ListFormatter):
    """Table formatter for visualizing tabulated data.
    """
    caption: Optional[str] = None

    def _init(self, show_detail: bool):
        self._table = Table(title=self.name, caption=self.caption)
        self._make_header(show_detail)

    def render(self, data: List[dict], show_detail: bool = False) -> None:
        self._build_table(data, show_detail)
        self._console.print(self._table)

    def get_renderable(self, data: List[dict], show_detail: bool = False) -> Table:
        self._build_table(data, show_detail)
        return self._table

    def _build_table(self, data: List[dict], show_detail: bool):
        self._init(show_detail)

        for d in data:
            info = [ self._substitute(get_value(d, f)) for f in self.fields ]
            if show_detail:
                info.extend([ self._substitute(get_value(d, f)) for f in self.extra_fields ])
            self._table.add_row(*info)

    def _make_header(self, show_detail: bool) -> None:
        for header in self.headers:
            self._table.add_column(header, **self._styling_map.get(header, {}))

        if show_detail:
            for extra_header in self.extra_headers:
                self._table.add_column(extra_header)


@dataclass
class PanelFormatter(ListFormatter):
    """Panel formatter for visualizing information detail in a panel.
    """
    subtitle: Optional[str] = None

    def render(self, data: Union[List[dict], dict], show_detail: bool = False) -> None:
        if not isinstance(data, list):
            data = [data]
        self._build_panel(data, show_detail)
        self._console.print(self._panel)

    def get_renderable(self, data: Union[List[dict], dict], show_detail: bool = False) -> Panel:
        if not isinstance(data, list):
            data = [data]
        self._build_panel(data, show_detail)
        return self._panel

    def _build_panel(self, data: List[dict], show_detail: bool):
        headers = self.headers + self.extra_headers if show_detail else self.headers
        table = Table(box=None, show_header=False)
        table.add_column('k', style='dim bold')
        table.add_column('v')

        for d in data:
            info = [ self._substitute(get_value(d, f)) for f in self.fields ]
            if show_detail:
                info.extend([ self._substitute(get_value(d, f)) for f in self.extra_fields ])
            for k, v in zip(headers, info):
                table.add_row(k, v)
        self._panel = Panel(table, title=self.name, subtitle=self.subtitle)


@dataclass
class Edge:
    """Edge of file traversal tree.
    """
    name: str
    size: int


def find_and_insert(parent: Tree, edges: List[Edge]) -> None:
    if not edges:
        return

    is_dir = len(edges) > 1
    folder_style_prefix = "[bold magenta]:open_file_folder: "

    match = [
        tree for tree in parent.children \
            if edges[0].name == str(tree.label).replace(folder_style_prefix, "").split(' ')[0]
    ]

    if match:
        tree = match[0]
    else:
        if is_dir:
            tree = parent.add(f"{folder_style_prefix}{edges[0].name}")
        else:
            text_filename = Text(edges[0].name, "green")
            file_size = edges[0].size
            text_filename.append(f" ({decimal(file_size)})", "blue")
            tree = parent.add(text_filename)

    find_and_insert(tree, edges[1:])


@dataclass
class TreeFormatter(Formatter):
    """Tree formatter for visualizing file tree.
    """
    def __post_init__(self):
        super().__post_init__()

    def render(self, data: List[dict]) -> None:
        self._build_renderable(data)
        self._console.print(self._panel)

    def get_renderable(self, data: List[dict]) -> Panel:
        self._build_renderable(data)
        return self._panel

    def _build_tree(self, data: List[dict]) -> Tree:
        root = Tree("/")
        paths = [ f"/{d['path'].lstrip()}" for d in data ]
        sizes = [ d['size'] for d in data]
        for path, size in zip(paths, sizes):
            edges = []
            parts = path.split("/")[1:]
            for i, part in enumerate(parts):
                file_size = size if i == len(parts) - 1 else None
                edges.append(Edge(part, file_size))
            find_and_insert(root, edges)
        return root

    def _build_renderable(self, data: List[dict]) -> None:
        root = self._build_tree(data)
        self._panel = Panel(root, title=self.name)


@dataclass
class JSONFormatter(Formatter):
    """JSON formatter for visualizing JSON data.
    """
    def render(self, data: dict) -> None:
        self._build_json(data)
        self._console.print(self._panel)

    def get_renderable(self, data: List[dict]) -> Panel:
        self._build_json(data)
        return self._panel

    def _build_json(self, data: dict) -> None:
        self._panel = Panel(JSON(json.dumps(data)), title=self.name)



# Copyright (C) 2022 FriendliAI

from typing import List

from pfcli.service.client.base import ClientService, T, safe_request
from pfcli.utils import paginated_get


class BillingClientService(ClientService):

    def list_prices(self,
                    year: int,
                    month: int,
                    day: int = None,
                    group_id: T = None,
                    project_id: T = None) -> List[dict]:
        params = {
            "group_id": str(group_id) if group_id else None,
            "project_id": str(project_id) if project_id else None,
            "year": year,
            "month": month,
            "day": day
        }
        get_response_dict = safe_request(self.list, err_prefix=f"Failed to get billing summary.")
        return paginated_get(
            get_response_dict,
            **params
        )



# Copyright (C) 2021 FriendliAI

"""PeriFlow Experiment CLI"""

from typing import Optional
from datetime import datetime
from dateutil.parser import parse

import typer

from pfcli.service import ServiceType
from pfcli.service.client import ExperimentClientService, ProjectExperimentClientService, build_client
from pfcli.service.formatter import PanelFormatter, TableFormatter
from pfcli.job import job_table
from pfcli.utils import datetime_to_pretty_str, timedelta_to_pretty_str

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False
)
table_formatter = TableFormatter(name="Experiments", fields=['name'], headers=['Name'])
panel_formatter = PanelFormatter(name="Overview", fields=['name'], headers=['Name'])


@app.command()
def create(
    name: str = typer.Option(
        ...,
        '--name',
        '-n',
        help="The name of experiment to create."
    )
):
    """Create a new experiment.
    """
    client: ProjectExperimentClientService = build_client(ServiceType.PROJECT_EXPERIMENT)
    experiment = client.create_experiment(name)
    panel_formatter.render([experiment])


@app.command()
def list():
    """List all experiments.
    """
    client: ProjectExperimentClientService = build_client(ServiceType.PROJECT_EXPERIMENT)
    experiments = client.list_experiments()
    table_formatter.render(experiments)


@app.command()
def view(
    name: str = typer.Argument(
        ...,
        help="The name of experiment to view details."
    ),
    tail: Optional[int] = typer.Option(
        None,
        "--tail",
        help="The number of job list to view at the tail"
    ),
    head: Optional[int] = typer.Option(
        None,
        "--head",
        help="The number of job list to view at the head"
    ),
):
    """Show a list of jobs in the experiment.
    """
    client: ExperimentClientService = build_client(ServiceType.EXPERIMENT)
    project_client: ProjectExperimentClientService = build_client(ServiceType.PROJECT_EXPERIMENT)
    experiment_id = project_client.get_id_by_name(name)
    jobs = client.list_jobs_in_experiment(experiment_id)

    for job in jobs:
        started_at = job.get("started_at")
        finished_at = job.get("finished_at")
        if started_at is not None:
            start = datetime_to_pretty_str(parse(job["started_at"]))
        else:
            start = None
        if started_at is not None and finished_at is not None:
            duration = timedelta_to_pretty_str(parse(started_at), parse(finished_at))
        elif started_at is not None and job["status"] == "running":
            start_time = parse(started_at)
            curr_time = datetime.now(start_time.tzinfo)
            duration = timedelta_to_pretty_str(start_time, curr_time)
        else:
            duration = None
        job['started_at'] = start
        job['duration'] = duration
        job['data_name'] = job["data_store"]['name'] if job["data_store"] is not None else None

    if tail is not None or head is not None:
        target_job_list = []
        if tail:
            target_job_list.extend(jobs[:tail])
        if head:
            target_job_list.extend(jobs[-head:]) 
    else:
        target_job_list = jobs

    job_table.render(target_job_list)


@app.command()
def edit(
    name: str = typer.Argument(
        ...,
        help="The name of experiment to update."
    ),
    new_name: str = typer.Option(
        ...,
        '--name',
        '-n',
        help="The new new of experiment"
    )
):
    """Edit experiment info.
    """
    client: ExperimentClientService = build_client(ServiceType.EXPERIMENT)
    project_client: ProjectExperimentClientService = build_client(ServiceType.PROJECT_EXPERIMENT)

    experiment_id = project_client.get_id_by_name(name)
    experiment = client.update_experiment_name(experiment_id, new_name)
    panel_formatter.render([experiment])


@app.command()
def delete(
    name: str = typer.Argument(
        ...,
        help="The name of experiment to delete."
    ),
    force: bool = typer.Option(
        False,
        '--force',
        '-f',
        help="Forcefully delete experiments and all jobs inside the experiment without confirmation prompt."
    )
):
    """Delete a experiment.
    """
    if not force:
        do_delete = typer.confirm(
            "!!! This action is VERY DESTRUCTIVE !!!\n"
            "Are your sure to delete experiment and ALL THE JOBS inside the experiment?"
        )
        if not do_delete:
            raise typer.Abort()
        really_do_delete = typer.confirm("STOP!!! Are you really sure to delete? This is the last confirmation.")
        if not really_do_delete:
            raise typer.Abort()

    client: ExperimentClientService = build_client(ServiceType.EXPERIMENT)
    project_client: ProjectExperimentClientService = build_client(ServiceType.PROJECT_EXPERIMENT)

    experiment_id = project_client.get_id_by_name(name)
    client.delete_experiment(experiment_id)

    typer.secho(f"Experiment ({name}) is deleted successfully!", fg=typer.colors.BLUE)



# Copyright (C) 2021 FriendliAI

"""PeriFlow YAML File Configuration Service"""

import os
import tempfile
from dataclasses import dataclass, field
from typing import (
    Any,
    List,
    Tuple,
    TypeVar,
    Optional,
    Union,
)

import yaml
import typer
from click import Choice
from jsonschema import Draft7Validator, ValidationError

from pfcli.service import (
    StorageType,
    JobType,
    ServiceType,
    CredType,
    cred_type_map_inv,
    storage_region_map,
)
from pfcli.service.client import (
    CredentialClientService,
    CredentialTypeClientService,
    ProjectCredentialClientService,
    JobTemplateClientService,
    build_client,
)
from pfcli.service.cloud import build_storage_helper
from pfcli.utils import get_default_editor, open_editor, secho_error_and_exit


DEFAULT_TEMPLATE_CONFIG = """\
# The name of experiment
experiment:

# The name of job
name:

# The name of vm type
vm:

# The number of GPU devices
num_devices:
"""

DATA_CONFIG = """
# Configure dataset
data:
  # The name of dataset
  name:
"""

DATA_MOUNT_CONFIG = """\
  # Path to mount your dataset volume
  mount_path:
"""


JOB_SETTING_CONFIG = """
# Configure your job!
job_setting:
"""

CUSTOM_JOB_SETTING_CONFIG = JOB_SETTING_CONFIG + """\
  type: custom

  # Docker config
  docker:
    # Docker image you want to use in the job
    image:
    # Bash shell command to run the job.
    #
    # NOTE: PeriFlow automatically sets the following environment variables for PyTorch DDP.
    #   - MASTER_ADDR: Address of rank 0 node.
    #   - WORLD_SIZE: The total number of GPUs participating in the task.
    #   - NODE_RANK: Index of the current node.
    #   - NPROC_PER_NODE: The number of processes in the current node.
    command:
"""

PRIVATE_DOCKER_IMG_CONFIG = """\
    credential_id:
"""

JOB_WORKSPACE_CONFIG = """\
  # Path to mount your workspace volume. If not specified, '/workspace' will be used by default.
  workspace:
    mount_path:
"""

PREDEFINED_JOB_SETTING_CONFIG = JOB_SETTING_CONFIG + """\
  type: predefined
"""

DIST_CONFIG = """
# Distributed training config
dist:
  dp_degree:
  pp_degree:
  mp_degree:
"""

CHECKPOINT_CONFIG = """
# Checkpoint config
checkpoint:
"""

INPUT_CHECKPOINT_CONFIG = """\
  input:
    # UUID of input checkpoint
    id:
"""

INPUT_CHECKPOINT_MOUNT_PATH = """\
    # Input checkpoint mount path
    mount_path:
"""

OUTPUT_CHECKPOINT_CONFIG = """\
  # Path to output checkpoint
  output_checkpoint_dir:
"""

PLUGIN_CONFIG = """
# Additional plugin for job monitoring and push notification
plugin:
"""

WANDB_PLUGIN_CONFIG = """\
  wandb:
    # W&B API key
    credential_id:
"""

SLACK_PLUGIN_CONFIG = """\
  slack:
    credential_id:
    channel:
"""


J = TypeVar('J', bound='JobConfigService')
D = TypeVar('D', bound='DataConfigService')
T = TypeVar('T', bound=Union[str, Tuple[Any, ...]])


class InteractiveConfigMixin:
    def start_interaction(self) -> None:
        raise NotImplementedError   # prama: no cover

    def render(self) -> T:
        raise NotImplementedError   # pragma: no cover


@dataclass
class JobConfigService(InteractiveConfigMixin):
    """Interface of job template configuration service
    """
    ready: bool = False
    use_data: bool = False
    use_input_checkpoint: bool = False
    use_wandb: bool = False
    use_slack: bool = False

    def _render_plugins(self) -> str:
        yaml_str = ""
        if any((self.use_wandb, self.use_slack)):
            yaml_str += PLUGIN_CONFIG
            if self.use_wandb:
                yaml_str += WANDB_PLUGIN_CONFIG
            if self.use_slack:
                yaml_str += SLACK_PLUGIN_CONFIG

        return yaml_str


@dataclass
class CustomJobConfigService(JobConfigService):
    """Custom job template configuration service
    """
    # TODO: Support artifact
    use_private_image: bool = False
    use_dist: bool = False
    use_output_checkpoint: bool = False

    def start_interaction(self):
        self.use_private_image = typer.confirm(
            "Will you use your private docker image? (You should provide a credential).", prompt_suffix="\n>> "
        )
        self.use_workspace = typer.confirm(
            "Do you want to run the job with the scripts in your local directory?", prompt_suffix="\n>> "
        )
        self.use_data = typer.confirm(
            "Will you use a dataset for the job?", prompt_suffix="\n>> "
        )
        self.use_input_checkpoint = typer.confirm(
            "Will you use an input checkpoint for the job?", prompt_suffix="\n>> "
        )
        self.use_output_checkpoint = typer.confirm(
            "Does your job generate model checkpoint files?", prompt_suffix="\n>> "
        )
        self.use_dist = typer.confirm(
            "Will you run distributed training job?", prompt_suffix="\n>> "
        )
        self.use_wandb = typer.confirm(
            "Will you use W&B monitoring for the job?", prompt_suffix="\n>> "
        )
        self.use_slack = typer.confirm(
            "Do you want to get a Slack notification for the job?", prompt_suffix="\n>> "
        )
        self.ready = True

    def render(self) -> str:
        assert self.ready

        yaml_str = DEFAULT_TEMPLATE_CONFIG
        yaml_str += CUSTOM_JOB_SETTING_CONFIG
        if self.use_private_image:
            yaml_str += PRIVATE_DOCKER_IMG_CONFIG
        if self.use_workspace:
            yaml_str += JOB_WORKSPACE_CONFIG
        if self.use_data:
            yaml_str += DATA_CONFIG
            yaml_str += DATA_MOUNT_CONFIG
        if any((self.use_input_checkpoint, self.use_output_checkpoint)):
            yaml_str += CHECKPOINT_CONFIG
            if self.use_input_checkpoint:
                yaml_str += INPUT_CHECKPOINT_CONFIG
                yaml_str += INPUT_CHECKPOINT_MOUNT_PATH
            if self.use_output_checkpoint:
                yaml_str += OUTPUT_CHECKPOINT_CONFIG
        if self.use_dist:
            yaml_str += DIST_CONFIG
        yaml_str += self._render_plugins()

        return yaml_str


@dataclass
class PredefinedJobConfigService(JobConfigService):
    """Predefined job template configuration service
    """
    model_name: Optional[str] = None
    model_config: Optional[dict] = None

    def start_interaction(self) -> None:
        job_template_client_service: JobTemplateClientService = build_client(ServiceType.JOB_TEMPLATE)

        template_names = job_template_client_service.list_job_template_names()
        self.model_name = typer.prompt(
            "Which job do you want to run? Choose one in the following catalog:\n",
            type=Choice(template_names),
            prompt_suffix="\n>> "
        )
        template = job_template_client_service.get_job_template_by_name(self.model_name)
        assert template is not None
        self.model_config = template["data_example"]

        self.use_data = typer.confirm(
            "Will you use dataset for the job?", prompt_suffix="\n>> "
        )
        self.use_input_checkpoint = typer.confirm(
            "Will you use input checkpoint for the job?", prompt_suffix="\n>> "
        )
        self.use_wandb = typer.confirm(
            "Will you use W&B monitoring for the job?", prompt_suffix="\n>> "
        )
        self.use_slack = typer.confirm(
            "Do you want to get slack notifaction for the job?", prompt_suffix="\n>> "
        )
        self.ready = True

    def render(self) -> str:
        assert self.ready

        yaml_str = DEFAULT_TEMPLATE_CONFIG
        yaml_str += PREDEFINED_JOB_SETTING_CONFIG
        yaml_str += f"  template_name: {self.model_name}\n"
        yaml_str += f"  model_config:\n"
        for k, v in self.model_config.items():
            yaml_str += f"    {k}: {v}\n"

        if self.use_data:
            yaml_str += DATA_CONFIG
        if self.use_input_checkpoint:
            yaml_str += CHECKPOINT_CONFIG
            yaml_str += INPUT_CHECKPOINT_CONFIG
        yaml_str += self._render_plugins()

        return yaml_str


@dataclass
class CredentialConfigService(InteractiveConfigMixin):
    """Credential configuration service
    """
    ready: bool = False
    name: Optional[str] = None
    cred_type: Optional[CredType] = None
    value: Optional[dict] = None

    def start_interaction(self) -> None:
        self.name = typer.prompt(
            "Enter the name of your new credential.", prompt_suffix="\n>> "
        )
        self.cred_type = typer.prompt(
            "What kind of credential do you want to create?\n",
            type=Choice([ e.value for e in CredType ]),
            prompt_suffix="\n>> "
        )
        cred_type_client: CredentialTypeClientService = build_client(ServiceType.CREDENTIAL_TYPE)
        schema = cred_type_client.get_schema_by_type(self.cred_type)
        properties: dict = schema['properties']
        self.value = {}
        typer.echo("Please fill in the following fields")
        for field, field_info in properties.items():
            field_info: dict
            field_info_str = "\n".join(f"    - {k}: {v}" for k, v in field_info.items())
            hide_input = True if "password" in field else False
            entered = typer.prompt(f"  {field}:\n{field_info_str}", prompt_suffix="\n  >> ", hide_input=hide_input)
            self.value[field] = entered

        self._validate_schema(schema)
        self.ready = True

    def start_interaction_for_update(self, credential_id: str) -> None:
        cred_client: CredentialClientService = build_client(ServiceType.CREDENTIAL)
        prev_cred = cred_client.get_credential(credential_id)

        self.name = typer.prompt(
            "Enter the NEW name of your credential. Press ENTER if you don't want to update this.\n"
            f"Current: {prev_cred['name']}",
            prompt_suffix="\n>> ",
            default=prev_cred['name'],
            show_default=False
        )
        self.cred_type = cred_type_map_inv[prev_cred['type']]
        cred_type_client: CredentialTypeClientService = build_client(ServiceType.CREDENTIAL_TYPE)
        schema = cred_type_client.get_schema_by_type(self.cred_type)
        properties: dict = schema['properties']
        self.value = {}
        typer.echo("Please fill in the following fields")
        for field, field_info in properties.items():
            field_info: dict
            field_info_str = "\n".join(f"    - {k}: {v}" for k, v in field_info.items())
            hide_input = True if "password" in field else False
            entered = typer.prompt(
                f"  {field} (Current: {prev_cred['value'][field]}):\n{field_info_str}",
                prompt_suffix="\n  >> ",
                default=prev_cred['value'][field],
                show_default=False,
                hide_input=hide_input
            )
            self.value[field] = entered

        self._validate_schema(schema)
        self.ready = True

    def _validate_schema(self, schema: dict) -> None:
        try:
            Draft7Validator(schema).validate(self.value)
        except ValidationError as exc:
            secho_error_and_exit(f"Format of credential value is invalid...! ({exc.message})")

    def render(self) -> Tuple[Any, ...]:
        assert self.ready
        return self.name, self.cred_type, self.value


@dataclass
class DataConfigService(InteractiveConfigMixin):
    ready: bool = False
    name: Optional[str] = None
    vendor: Optional[StorageType] = None
    region: Optional[str] = None
    storage_name: Optional[str] = None
    credential_id: Optional[str] = None
    metadata: Optional[dict] = field(default_factory=dict)
    files: Optional[List[dict]] = field(default_factory=list)

    def _list_available_credentials(self, vendor_type: StorageType) -> List[dict]:
        cred_type: CredType = CredType(vendor_type.value)
        project_cred_client: ProjectCredentialClientService = build_client(ServiceType.PROJECT_CREDENTIAL)

        creds = project_cred_client.list_credentials(cred_type=cred_type)

        return creds

    def _get_credential(self) -> dict:
        client: CredentialClientService = build_client(ServiceType.CREDENTIAL)
        return client.get_credential(self.credential_id)['value']

    def start_interaction_common(self) -> None:
        self.name = typer.prompt(
            "Enter the name of your new dataset.", prompt_suffix="\n>> ",
        )
        self.vendor = typer.prompt(
            "Enter the cloud vendor where your dataset is uploaded.",
            type=Choice([e.value for e in StorageType]),
            prompt_suffix="\n>> "
        )
        self.region = typer.prompt(
            "Enter the region of cloud storage where your dataset is uploaded.",
            type=Choice(storage_region_map[self.vendor]),
            prompt_suffix="\n>> "
        )
        self.storage_name = typer.prompt(
            "Enter the storage name where your dataset is uploaded.", prompt_suffix="\n>> "
        )
        available_creds = self._list_available_credentials(self.vendor)
        cloud_cred_options = "\n".join(f"  - {cred['id']}: {cred['name']}" for cred in available_creds)
        self.credential_id = typer.prompt(
            "Enter credential UUID to access your cloud storage. "
            f"Your available credentials for cloud storages are:\n{cloud_cred_options}",
            type=Choice([ cred['id'] for cred in available_creds ]),
            show_choices=False,
            prompt_suffix="\n>> "
        )
        credential_value = self._get_credential()
        storage_helper = build_storage_helper(self.vendor, credential_value)
        self.files = storage_helper.list_storage_files(self.storage_name)

    def render(self) -> Tuple[Any, ...]:
        assert self.ready
        return self.name, self.vendor, self.region, self.storage_name, self.credential_id, self.metadata, self.files


@dataclass
class PredefinedDataConfigService(DataConfigService):
    model_name: Optional[str] = None

    def start_interaction(self) -> None:
        self.start_interaction_common()

        # Configure metdata
        job_template_client_service: JobTemplateClientService = build_client(ServiceType.JOB_TEMPLATE)
        template_names = job_template_client_service.list_job_template_names()
        self.model_name = typer.prompt(
            "Which job would you like to use this datastore? Choose one in the following catalog:\n",
            type=Choice(template_names),
            prompt_suffix="\n>> "
        )
        template = job_template_client_service.get_job_template_by_name(self.model_name)
        assert template is not None

        schema = template["data_store_template"]["metadata_schema"]
        properties: dict = schema['properties']
        self.metadata = {}
        typer.echo("Please fill in the following fields (NOTE: Enter comma-separated string for array values)")
        for field, field_info in properties.items():
            field_info: dict
            field_info_str = "\n".join(f"    - {k}: {v}" for k, v in field_info.items())
            entered = typer.prompt(f"  {field}:\n{field_info_str}", prompt_suffix="\n  >> ")
            if field_info['type'] == 'array':
                entered = entered.split(',')
            self.metadata[field] = entered

        try:
            Draft7Validator(schema).validate(self.metadata)
        except ValidationError as exc:
            secho_error_and_exit(f"Format of credential value is invalid...! ({exc.message})")

        self.ready = True


@dataclass
class CustomDataConfigService(DataConfigService):
    def start_interaction(self) -> None:
        self.start_interaction_common()

        # Configure metdata
        exist_metadata = typer.confirm(
            "[Optional] Do you want to add metadata for your datastore? "
            "You can use this metadata in PeriFlow serving service.",
        )
        if exist_metadata:
            editor = typer.prompt(
                "Editor will be opened in the current terminal to edit the metadata. "
                "The metadata should be described in YAML format.\n"
                f"Your default editor is '{get_default_editor()}'. "
                "If you want to use another editor, enter the name of your preferred editor.",
                default=get_default_editor(),
                prompt_suffix="\n>> "
            )
            with tempfile.TemporaryDirectory() as dir:
                path = os.path.join(dir, 'metadata.yaml')
                open_editor(path, editor)
                try:
                    with open(path, 'r', encoding='utf-8') as f:
                        self.metadata = yaml.safe_load(f)
                except yaml.YAMLError as exc:
                    secho_error_and_exit(f"Error occurred while parsing metadata file... {exc}")
        self.ready = True


def build_job_configurator(job_type: str) -> J:
    if job_type == "custom":
        configurator = CustomJobConfigService()
    elif job_type == "predefined":
        configurator = PredefinedJobConfigService()
    else:
        secho_error_and_exit("Invalid job type...!")

    configurator.start_interaction()
    return configurator


def build_data_configurator(job_type: JobType) -> D:
    if job_type == JobType.CUSTOM:
        configurator = CustomDataConfigService()
    elif job_type == JobType.PREDEFINED:
        configurator = PredefinedDataConfigService()
    else:
        secho_error_and_exit("Invalid job type...!")

    configurator.start_interaction()
    return configurator



# Copyright (C) 2022 FriendliAI

"""PeriFlow CheckpointClient Service"""

from typing import List, Optional

import requests
from requests.models import Response

from pfcli.service import StorageType
from pfcli.service.auth import auto_token_refresh, get_auth_header
from pfcli.service.client.base import ClientService, T, safe_request


class CheckpointClientService(ClientService):
    def get_checkpoint(self, checkpoint_id: T) -> dict:
        response = safe_request(self.retrieve, err_prefix="Failed to get info of checkpoint")(
            pk=checkpoint_id
        )
        return response.json()

    def update_checkpoint(self,
                          checkpoint_id: T,
                          *,
                          vendor: Optional[StorageType] = None,
                          region: Optional[str] = None,
                          credential_id: Optional[str] = None,
                          iteration: Optional[int] = None,
                          storage_name: Optional[str] = None,
                          files: Optional[List[dict]] = None,
                          dist_config: Optional[dict] = None,
                          data_config: Optional[dict] = None,
                          job_setting_config: Optional[dict] = None) -> dict:
        request_data = {}
        if vendor is not None:
            request_data['vendor'] = vendor
        if region is not None:
            request_data['region'] = region
        if credential_id is not None:
            request_data['credential_id'] = credential_id
        if iteration is not None:
            request_data['iteration'] = iteration
        if storage_name is not None:
            request_data['storage_name'] = storage_name
        if files is not None:
            request_data['files'] = files
        if dist_config is not None:
            request_data['dist_json'] = dist_config
        if data_config is not None:
            request_data['data_json'] = data_config
        if job_setting_config is not None:
            request_data['job_setting_json'] = job_setting_config

        response = safe_request(self.partial_update, err_prefix="Cannot update checkpoint.")(
            pk=checkpoint_id,
            json=request_data
        )
        return response.json()

    def delete_checkpoint(self, checkpoint_id: T) -> None:
        response = safe_request(self.delete, err_prefix="Failed to delete checkpoint.")(
            pk=checkpoint_id
        )
        return response

    @auto_token_refresh
    def download(self, checkpoint_id: T) -> Response:
        response = safe_request(self.retrieve, err_prefix="Failed to get info of checkpoint.")(
            pk=checkpoint_id
        )
        model_form_id = response.json()['forms'][0]['id']

        url_template = self.url_template.copy()
        url_template.replace_path('model_forms/$model_form_id/download/')
        return requests.get(
            url_template.render(model_form_id=model_form_id, **self.url_kwargs),
            headers=get_auth_header(),
        )

    def get_checkpoint_download_urls(self, checkpoint_id: T) -> List[dict]:
        response = safe_request(self.download, err_prefix="Failed to get download URLs of checkpoint files.")(
            checkpoint_id=checkpoint_id
        )
        return response.json()['files']



# Copyright (C) 2021 FriendliAI

"""PeriFlow VM CLI"""

from typing import Optional

import typer

from pfcli.service import CloudType, ServiceType
from pfcli.service.client import (
    GroupVMConfigClientService,
    ProjectVMQuotaClientService,
    build_client
)
from pfcli.service.client.project import ProjectVMConfigClientService
from pfcli.service.formatter import TableFormatter
from pfcli.utils import validate_cloud_region


app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False
)

formatter = TableFormatter(
    name="VM Instances",
    fields=[
        'vm_config_type.code',
        'vm_config_type.vm_instance_type.vendor',
        'vm_config_type.vm_instance_type.region',
        'vm_config_type.vm_instance_type.device_type',
        'quota'
    ],
    headers=['VM', 'Cloud', 'Region', 'Device', 'Quota (Available / Total)']
)


@app.command("list")
def list(
    cloud: Optional[CloudType] = typer.Option(
        None,
        '--cloud',
        '-c',
        help="Filter list by cloud vendor."
    ),
    region: Optional[str] = typer.Option(
        None,
        '--region',
        '-r',
        help="Filter list by region."
    ),
    device_type: Optional[str] = typer.Option(
        None,
        '--device-type',
        '-d',
        help="Filter list by device type."
    )
):
    """List all VM quota information.
    """
    if cloud is not None and region is not None:
        validate_cloud_region(cloud, region)

    vm_quota_client: ProjectVMQuotaClientService = build_client(ServiceType.PROJECT_VM_QUOTA)
    vm_config_client: ProjectVMConfigClientService = build_client(ServiceType.PROJECT_VM_CONFIG)
    group_vm_config_client: GroupVMConfigClientService = build_client(ServiceType.GROUP_VM_CONFIG)

    vm_dict_list = vm_quota_client.list_vm_quotas(vendor=cloud, region=region, device_type=device_type)
    vm_id_map = group_vm_config_client.get_vm_config_id_map()
    for vm_dict in vm_dict_list:
        vm_instance_name = vm_dict['vm_config_type']['code']
        active_vm_count = vm_config_client.get_active_vm_count(vm_id_map[vm_instance_name])
        vm_dict['quota'] = f"{vm_dict['quota'] - active_vm_count} / {vm_dict['quota']}"

    formatter.render(vm_dict_list)



# Copyright (C) 2022 FriendliAI

"""PeriFlow ProjectClient Service"""


import json
import uuid
from pathlib import Path
from string import Template
from typing import List, Optional
from requests import HTTPError

from rich.filesize import decimal

from pfcli.service import CloudType, CredType, LockType, StorageType, cred_type_map, storage_type_map
from pfcli.service.client.base import ClientService, ProjectRequestMixin, T, safe_request
from pfcli.utils import get_workspace_files, paginated_get, secho_error_and_exit, validate_storage_region, zip_dir


class ProjectClientService(ClientService):
    def get_project(self, pf_project_id: uuid.UUID) -> dict:
        response = safe_request(self.retrieve, err_prefix="Failed to get a project.")(
            pk=pf_project_id
        )
        return response.json()

    def check_project_membership(self, pf_project_id: uuid.UUID) -> bool:
        try:
            self.retrieve(pf_project_id)
        except HTTPError:
            return False
        else:
            return True

    def delete_project(self, pf_project_id: uuid.UUID) -> None:
        safe_request(self.delete, err_prefix="Failed to delete a project.")(
            pk=pf_project_id
        )

    def list_users(self, pf_project_id: str) -> List[dict]:
        get_response_dict = safe_request(self.list, err_prefix="Failed to list users in the current project")
        return paginated_get(get_response_dict, path=f"{pf_project_id}/pf_user")


class ProjectExperimentClientService(ClientService, ProjectRequestMixin):
    def __init__(self, template: Template, **kwargs):
        self.initialize_project()
        super().__init__(template, project_id=self.project_id, **kwargs)

    def list_experiments(self) -> List[dict]:
        response = safe_request(self.list, err_prefix="Failed to list experiments.")()
        return response.json()

    def get_id_by_name(self, name: str) -> Optional[T]:
        response = safe_request(self.list, err_prefix="Failed to get experiment info.")()
        for experiment in response.json():
            if experiment['name'] == name:
                return experiment['id']
        return None

    def create_experiment(self, name: str) -> dict:
        response = safe_request(self.post, err_prefix="Failed to post new experiment.")(
            data={"name": name}
        )
        return response.json()


class ProjectJobClientService(ClientService, ProjectRequestMixin):
    def __init__(self, template: Template, **kwargs):
        self.initialize_project()
        super().__init__(template, project_id=self.project_id, **kwargs)

    def list_jobs(self) -> List[dict]:
        return paginated_get(safe_request(self.list, err_prefix="Failed to list jobs in project."))

    def run_job(self, config: dict, workspace_dir: Optional[Path]) -> dict:
        job_request = safe_request(self.post, err_prefix="Failed to run job.")
        if workspace_dir is not None:
            workspace_dir = workspace_dir.resolve()
            workspace_files = get_workspace_files(workspace_dir)
            workspace_size = sum(f.stat().st_size for f in workspace_files)
            if workspace_size <= 0 or workspace_size > 100 * 1024 * 1024:
                secho_error_and_exit(
                    f"Workspace directory size ({decimal(workspace_size)}) should be 0 < size <= 100MB."
                )
            workspace_zip = Path(workspace_dir.parent / (workspace_dir.name + ".zip"))
            with zip_dir(workspace_dir, workspace_files, workspace_zip) as zip_file:
                files = {'workspace_zip': ('workspace.zip', zip_file)}
                response = job_request(
                    data={"data": json.dumps(config)}, files=files
                )
        else:
            response = job_request(json=config)
        return response.json()


class ProjectDataClientService(ClientService, ProjectRequestMixin):
    def __init__(self, template: Template, **kwargs):
        self.initialize_project()
        super().__init__(template, project_id=self.project_id, **kwargs)

    def list_datastores(self) -> List[dict]:
        response = safe_request(self.list, err_prefix="Failed to list dataset info.")()
        return response.json()

    def get_id_by_name(self, name: str) -> Optional[T]:
        datastores = self.list_datastores()
        for datastore in datastores:
            if datastore['name'] == name:
                return datastore['id']
        return None

    def create_datastore(self,
                         name: str,
                         vendor: StorageType,
                         region: str,
                         storage_name: str,
                         credential_id: Optional[T],
                         metadata: dict,
                         files: List[dict],
                         active: bool) -> dict:
        validate_storage_region(vendor, region)

        vendor_name = storage_type_map[vendor]
        request_data = {
            "name": name,
            "vendor": vendor_name,
            "region": region,
            "storage_name": storage_name,
            "credential_id": credential_id,
            "metadata": metadata,
            "files": files,
            "active": active,
        }
        response = safe_request(self.post, err_prefix="Failed to create a new datastore.")(
            json=request_data
        )
        return response.json()


class ProjectVMQuotaClientService(ClientService, ProjectRequestMixin):
    def __init__(self, template: Template, **kwargs):
        self.initialize_project()
        super().__init__(template, project_id=self.project_id, **kwargs)

    def list_vm_quotas(self,
                       vendor: Optional[CloudType] = None,
                       region: Optional[str] = None,
                       device_type: Optional[str] = None) -> Optional[List[dict]]:
        response = safe_request(self.list, err_prefix="Failed to list VM quota info.")()
        vm_dict_list = response.json()
        if vendor is not None:
            vm_dict_list = list(filter(lambda info: info['vm_config_type']['vm_instance_type']['vendor'] == vendor, vm_dict_list))
        if region is not None:
            vm_dict_list = list(filter(lambda info: info['vm_config_type']['vm_instance_type']['region'] == region, vm_dict_list))
        if device_type is not None:
            vm_dict_list = list(filter(lambda info: info['vm_config_type']['vm_instance_type']['device_type'] == device_type, vm_dict_list))
        return vm_dict_list


class ProjectCredentialClientService(ClientService, ProjectRequestMixin):
    def __init__(self, template: Template, **kwargs):
        self.initialize_project()
        super().__init__(template, project_id=self.project_id, **kwargs)

    def list_credentials(self, cred_type: CredType) -> List[dict]:
        type_name = cred_type_map[cred_type]
        response = safe_request(self.list, err_prefix=f"Failed to list credential for {cred_type}.")(
            params={"type": type_name}
        )
        return response.json()

    def create_credential(self,
                          cred_type: CredType,
                          name: str,
                          type_version: int,
                          value: dict) -> dict:
        type_name = cred_type_map[cred_type]
        request_data = {
            "type": type_name,
            "name": name,
            "type_version": type_version,
            "value": value
        }
        response = safe_request(self.post, err_prefix="Failed to create user credential.")(
            json=request_data
        )
        return response.json()


class ProjectVMConfigClientService(ClientService, ProjectRequestMixin):
    def __init__(self, template: Template, **kwargs):
        self.initialize_project()
        super().__init__(template, project_id=self.project_id, **kwargs)

    def list_vm_locks(self, vm_config_id: T, lock_type: LockType) -> List[dict]:
        response = safe_request(self.list, err_prefix="Failed to inspect locked VMs.")(
            path=f"{vm_config_id}/vm_lock/",
            params={"lock_type": lock_type}
        )
        return response.json()

    def get_active_vm_count(self, vm_config_id: T) -> int:
        vm_locks = self.list_vm_locks(vm_config_id, LockType.ACTIVE)
        return len(vm_locks)



# Copyright (C) 2022 FriendliAI

"""PeriFlow Client Service"""

import copy
import requests
from dataclasses import dataclass
from functools import partial, wraps
from requests.models import Response
from string import Template
from typing import Callable, Optional, TypeVar, Union
from urllib.parse import urljoin, urlparse

import uuid

from pfcli.context import get_current_group_id, get_current_project_id
from pfcli.service.auth import auto_token_refresh, get_auth_header
from pfcli.utils import decode_http_err, get_auth_uri, secho_error_and_exit


T = TypeVar('T', bound=Union[int, str, uuid.UUID])


def safe_request(func: Optional[Callable[..., Response]] = None, *, err_prefix: str = "") -> Callable[..., Response]:
    if err_prefix:
        err_prefix = err_prefix.rstrip() + "\n"

    if func is None:
        return partial(safe_request, err_prefix=err_prefix)

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except requests.HTTPError as exc:
            secho_error_and_exit(err_prefix + decode_http_err(exc))

    return wrapper


@dataclass
class URLTemplate:
    pattern: Template

    def render(self, pk: Optional[T] = None, path: Optional[str] = None, **kwargs) -> str:
        """render URLTemplate

        Args:
            pk: primary key
            path: additional path to attach
        """
        if pk is None and path is None:
            return self.pattern.substitute(**kwargs)

        pattern = copy.deepcopy(self.pattern)
        need_trailing_slash = pattern.template.endswith('/')

        if pk is not None:
            pattern.template = urljoin(pattern.template + '/', str(pk))
            if need_trailing_slash:
                pattern.template += '/'

        if path is not None:
            pattern.template = urljoin(pattern.template + '/', path.rstrip('/'))
            if need_trailing_slash:
                pattern.template += '/'

        return pattern.substitute(**kwargs)

    def get_base_url(self) -> str:
        result = urlparse(self.pattern.template)
        return f"{result.scheme}://{result.hostname}"

    def attach_pattern(self, pattern: str) -> None:
        self.pattern.template = urljoin(self.pattern.template + '/', pattern)

    def replace_path(self, path: str):
        result = urlparse(self.pattern.template)
        result = result._replace(path=path)
        self.pattern.template = result.geturl()

    def copy(self) -> "URLTemplate":
        return URLTemplate(pattern=Template(self.pattern.template))


class ClientService:
    def __init__(self, template: Template, **kwargs):
        self.url_template = URLTemplate(template)
        self.url_kwargs = kwargs

    @auto_token_refresh
    def list(self, path: Optional[str] = None, **kwargs) -> Response:
        return requests.get(
            self.url_template.render(path=path, **self.url_kwargs),
            headers=get_auth_header(),
            **kwargs
        )

    @auto_token_refresh
    def retrieve(self, pk: T, path: Optional[str] = None, **kwargs) -> Response:
        return requests.get(
            self.url_template.render(pk=pk, path=path, **self.url_kwargs),
            headers=get_auth_header(),
            **kwargs
        )

    @auto_token_refresh
    def post(self, path: Optional[str] = None, **kwargs) -> Response:
        return requests.post(
            self.url_template.render(path=path, **self.url_kwargs),
            headers=get_auth_header(),
            **kwargs
        )

    @auto_token_refresh
    def partial_update(self, pk: T, path: Optional[str] = None, **kwargs) -> Response:
        return requests.patch(
            self.url_template.render(pk=pk, path=path, **self.url_kwargs),
            headers=get_auth_header(),
            **kwargs
        )

    @auto_token_refresh
    def delete(self, pk: T, path: Optional[str] = None, **kwargs) -> Response:
        return requests.delete(
            self.url_template.render(pk=pk, path=path, **self.url_kwargs),
            headers=get_auth_header(),
            **kwargs
        )

    @auto_token_refresh
    def update(self, pk: T, path: Optional[str] = None, **kwargs) -> Response:
        return requests.put(
            self.url_template.render(pk=pk, path=path, **self.url_kwargs),
            headers=get_auth_header(),
            **kwargs
        )
    
    def bare_post(self, path: Optional[str] = None, **kwargs) -> Response:
        r = requests.post(
            self.url_template.render(path=path, **self.url_kwargs),
            **kwargs
        )
        r.raise_for_status()
        return r


class UserRequestMixin:
    user_id: uuid.UUID

    @auto_token_refresh
    def _userinfo(self) -> Response:
        return requests.get(get_auth_uri("oauth2/userinfo"), headers=get_auth_header())

    def get_current_userinfo(self) -> dict:
        response = safe_request(self._userinfo, err_prefix="Failed to get userinfo.")()
        return response.json()

    def get_current_user_id(self) -> uuid.UUID:
        userinfo = self.get_current_userinfo()
        return uuid.UUID(userinfo['sub'].split('|')[1])

    def initialize_user(self):
        self.user_id = self.get_current_user_id()


class GroupRequestMixin:
    group_id: uuid.UUID

    def initialize_group(self):
        group_id = get_current_group_id()
        if group_id is None:
            secho_error_and_exit("Organization is not set.")
        self.group_id = group_id


class ProjectRequestMixin:
    project_id: uuid.UUID

    def initialize_project(self):
        project_id = get_current_project_id()
        if project_id is None:
            secho_error_and_exit("Project is not set.")
        self.project_id = project_id



# Copyright (C) 2022 friendli.ai
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
from setuptools import setup, find_packages


def read(fname: str) -> str:
    return open(os.path.join(os.path.dirname(__file__), fname), encoding='utf-8').read()


def read_version() -> str:
    return read("VERSION").strip()


def read_readme() -> str:
    return read("README.md")


COMMON_DEPS = [
    "requests>=2.26.0",
    "tabulate>=0.8.0",
    "websockets>=10.1",
    "PyYaml>=6.0",
    "ruamel.yaml>=0.17.21",
    "typer>=0.4.0",
    "rich>=12.2.0",
    "jsonschema>=4.4.0",
    "boto3>=1.20.*",
    "botocore>=1.23.*",
    "tqdm>=4.64.0",
    "azure-mgmt-storage>=19.0.*",
    "azure-storage-blob>=12.9.*",
    "pathspec>=0.9.0",
]

TEST_DEPS = [
    "coverage==5.5",
    "pytest==6.2.4",
    "pytest-asyncio==0.15.1",
    "pytest-cov==2.11.1",
    "pytest-benchmark==3.4.1",
    "pytest-lazy-fixture==0.6.3",
    "requests-mock>=1.9.3",
]

setup(
    name='periflow-cli',
    version=read_version(),
    long_description=read_readme(),
    long_description_content_type='text/markdown',
    description='PeriFlow CLI',
    author='FriendliAI',
    license="Apache License 2.0",
    url="https://github.com/friendliai/periflow-cli",
    packages=find_packages(include=['pfcli', 'pfcli.*']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Natural Language :: English",
    ],
    entry_points={
        "console_scripts": [
            "pf = pfcli:app",
        ]
    },
    python_requires='>=3.6',
    include_package_data=True,
    install_requires=COMMON_DEPS,
    extras_require={
        "test": TEST_DEPS
    }
)



# Copyright (C) 2022 FriendliAI

"""PeriFlow DataClient Service"""


from pathlib import Path
from typing import Dict, List, Optional

from pfcli.service import StorageType, storage_type_map_inv
from pfcli.service.client.base import ClientService, T, safe_request
from pfcli.utils import secho_error_and_exit, validate_storage_region


class DataClientService(ClientService):
    def get_datastore(self, datastore_id: T) -> dict:
        response = safe_request(self.retrieve, err_prefix=f"Datastore ({datastore_id}) is not found.")(
            pk=datastore_id
        )
        return response.json()

    def update_datastore(self,
                         datastore_id: T,
                         *,
                         name: Optional[str] = None,
                         vendor: Optional[StorageType] = None,
                         region: Optional[str] = None,
                         storage_name: Optional[str] = None,
                         credential_id: Optional[str] = None,
                         metadata: Optional[dict] = None,
                         files: Optional[List[dict]] = None,
                         active: Optional[bool] = None) -> dict:
        # Valdiate region
        if vendor is not None or region is not None:
            prev_info = self.get_datastore(datastore_id)
            validate_storage_region(
                vendor or storage_type_map_inv[prev_info['vendor']],
                region or prev_info['region']
            )

        request_data = {}
        if name is not None:
            request_data['name'] = name
        if vendor is not None:
            request_data['vendor'] = vendor
        if region is not None:
            request_data['region'] = region
        if storage_name is not None:
            request_data['storage_name'] = storage_name
        if credential_id is not None:
            request_data['credential_id'] = credential_id
        if metadata is not None:
            request_data['metadata'] = metadata
        if files is not None:
            request_data['files'] = files
        if active is not None:
            request_data['active'] = active
        response = safe_request(self.partial_update, err_prefix="Failed to update datastore.")(
            pk=datastore_id,
            json=request_data
        )
        return response.json()

    def delete_datastore(self, datastore_id: T) -> None:
        safe_request(self.delete, err_prefix="Failed to delete datastore")(
            pk=datastore_id
        )

    def get_upload_urls(self, datastore_id: T, src_path: Path, expand: bool) -> List[Dict[str, str]]:
        if src_path.is_file():
            paths = [str(src_path.name)]
        else:
            paths = list(src_path.rglob('*'))
            rel_path = src_path if expand else src_path.parent
            paths = [str(f.relative_to(rel_path)) for f in paths if f.is_file()]
        if len(paths) == 0:
            secho_error_and_exit(f"No file exists in this path ({src_path})")

        response = safe_request(self.post, err_prefix="Failed to get presigned URLs.")(
            path=f"{datastore_id}/upload/",
            json={"paths": paths}
        )
        return response.json()



# Copyright (C) 2022 FriendliAI

"""PeriFlow GroupClient Service"""

import json
import uuid
from string import Template
from typing import Dict, List, Optional

from pfcli.service import CheckpointCategory, ModelFormCategory, StorageType
from pfcli.service.client.base import (
    ClientService,
    GroupRequestMixin,
    ProjectRequestMixin,
    T,
    UserRequestMixin,
    safe_request
)
from pfcli.utils import paginated_get, validate_storage_region


class GroupClientService(ClientService):
    def create_group(self, name: str) -> dict:
        response = safe_request(self.post, err_prefix="Failed to post an organization.")(
            data=json.dumps({"name": name, "hosting_type": "hosted"})
        )
        return response.json()

    def get_group(self, pf_group_id: uuid.UUID) -> dict:
        response = safe_request(self.retrieve, err_prefix="Failed to get an organization.")(
            pk=pf_group_id
        )
        return response.json()

    def invite_to_group(self, pf_group_id: uuid.UUID, email: str) -> None:
        safe_request(self.post, err_prefix="Failed to send invitation")(
            path=f"{pf_group_id}/invite",
            json={
                "email": email,
                "msg": ""
            }
        )

    def accept_invite(self, token: str) -> None:
        safe_request(self.post, err_prefix="Invalid code... Please Try again.")(
            path="invite/confirm",
            json={
                "email_token": token
            }
        )

    def get_user(self, pf_group_id: uuid.UUID, username: str) -> dict:
        get_response_dict = safe_request(self.list, err_prefix="Failed to get user in organization")
        return paginated_get(get_response_dict, path=f"{pf_group_id}/pf_user", search=username)

    def list_users(self, pf_group_id: uuid.UUID) -> List[dict]:
        get_response_dict = safe_request(self.list, err_prefix="Failed to list users in organization")
        return paginated_get(get_response_dict, path=f"{pf_group_id}/pf_user")


class GroupProjectClientService(ClientService, GroupRequestMixin):
    def __init__(self, template: Template, **kwargs):
        self.initialize_group()
        super().__init__(template, pf_group_id=self.group_id, **kwargs)

    def create_project(self, name: str) -> dict:
        response = safe_request(self.post, err_prefix="Failed to post a project.")(
            data=json.dumps({"name": name})
        )
        return response.json()

    def list_projects(self) -> List[dict]:
        get_response_dict = safe_request(self.list, err_prefix="Failed to list projects.")
        return paginated_get(get_response_dict)


class GroupVMConfigClientService(ClientService, GroupRequestMixin):
    def __init__(self, template: Template, **kwargs):
        self.initialize_group()
        super().__init__(template, group_id=self.group_id, **kwargs)

    def list_vm_configs(self) -> List[dict]:
        response = safe_request(self.list, err_prefix="Failed to list available VM list.")()
        return response.json()

    def get_vm_config_id_map(self) -> Dict[str, T]:
        id_map = {}
        for vm_config in self.list_vm_configs():
            id_map[vm_config['vm_config_type']['vm_instance_type']['code']] = vm_config['id']
        return id_map

    def get_id_by_name(self, name: str) -> Optional[T]:
        for vm_config in self.list_vm_configs():
            if vm_config['vm_config_type']['vm_instance_type']['code'] == name:
                return vm_config['id']
        return None


class GroupProjectCheckpointClientService(ClientService, UserRequestMixin, GroupRequestMixin, ProjectRequestMixin):
    def __init__(self, template: Template, **kwargs):
        self.initialize_user()
        self.initialize_group()
        self.initialize_project()
        super().__init__(
            template,
            group_id=self.group_id,
            project_id=self.project_id,
            **kwargs,
        )

    def list_checkpoints(self, category: Optional[CheckpointCategory]) -> List[dict]:
        request_data = {}
        if category is not None:
            request_data['category'] = category.value

        get_response_dict = safe_request(self.list, err_prefix="Failed to list checkpoints.")
        return paginated_get(get_response_dict, **request_data)

    def create_checkpoint(self,
                          name: str,
                          model_form_category: ModelFormCategory,
                          vendor: StorageType,
                          region: str,
                          credential_id: str,
                          iteration: int,
                          storage_name: str,
                          files: List[dict],
                          dist_config: dict,
                          data_config: dict,
                          job_setting_config: Optional[dict]) -> dict:
        validate_storage_region(vendor, region)

        request_data = {
            "job_id": None,
            "name": name,
            "attributes": {
                "data_json": data_config,
                "job_setting_json": job_setting_config,
            },
            "user_id": str(self.user_id),
            "credential_id": credential_id if credential_id else None,
            "model_category": "USER",
            "form_category": model_form_category.value,
            "dist_json": dist_config,
            "vendor": vendor,
            "region": region,
            "storage_name": storage_name,
            "iteration": iteration,
            "files": files
        }

        response = safe_request(self.post, err_prefix="Failed to post checkpoint.")(
            json=request_data
        )
        return response.json()
